# coding=utf-8
from __future__ import print_function

import functools
import pickle

from suanpan import path

_load = functools.partial(pickle.load, fix_imports=True, encoding="latin1")
_loads = functools.partial(pickle.loads, fix_imports=True, encoding="latin1")
_dump = functools.partial(pickle.dump, fix_imports=True)
_dumps = functools.partial(pickle.dumps, fix_imports=True)


def _loadf(file, *args, **kwargs):
    encoding = kwargs.pop("encoding", "utf-8")
    with open(file, "w", encoding=encoding) as file:
        return _load(file, *args, **kwargs)


def _dumpf(obj, file, *args, **kwargs):
    encoding = kwargs.pop("encoding", "utf-8")
    path.safeMkdirsForFile(file)
    with open(file, "w", encoding=encoding) as file:
        return _dump(obj, file, *args, **kwargs)


def load(file, *args, **kwargs):
    _l = _loadf if isinstance(file, str) else _load
    return _l(file, *args, **kwargs)


def dump(obj, file, *args, **kwargs):
    _d = _dumpf if isinstance(file, str) else _dump
    return _d(obj, file, *args, **kwargs)


loads = _loads
dumps = _dumps



# coding=utf-8
from __future__ import division, print_function

import os
import zipfile

from suanpan import path
from suanpan.log import logger


class LocalStorage(object):

    DEFAULT_IGNORE_KEYWORDS = {"__MACOSX", ".DS_Store"}

    def __init__(self, localTempStore="/tmp", **kwargs):
        self.delimiter = os.sep
        self.tempStore = localTempStore

    def download(self, name, path):
        return self.pathJoin(self.tempStore, path)

    def upload(self, name, path):
        return path

    def walk(self, folderName):
        root = self.pathJoin(self.tempStore, folderName)
        return os.walk(root)

    def listAll(self, folderName):
        root = self.pathJoin(self.tempStore, folderName)
        return (self.pathJoin(root, p) for p in os.listdir(root))

    def listFolders(self, folderName):
        return (p for p in self.listAll(folderName) if os.path.isdir(p))

    def listFiles(self, folderName):
        return (p for p in self.listAll(folderName) if os.path.isfile(p))

    def isFile(self, objectName):
        file = self.pathJoin(self.tempStore, objectName)
        return os.path.isfile(file)

    def isFolder(self, folderName):
        folder = self.pathJoin(self.tempStore, folderName)
        return os.path.isdir(folder)

    def getPathInTempStore(self, path, tempStore=None):
        tempStore = tempStore or self.tempStore
        return self.pathJoin(tempStore, path)

    def pathJoin(self, *paths, **kwargs):
        return os.path.join(*paths)

    def relativePath(self, path, base):
        base = base if base.endswith(self.delimiter) else base + self.delimiter
        return path[len(base) :] if path.startswith(base) else path

    def compress(self, zipFilePath, path, ignore=DEFAULT_IGNORE_KEYWORDS):
        compressFunc = self.compressFolder if os.path.isdir(path) else self.compressFile
        return compressFunc(zipFilePath, path)

    def compressFolder(self, zipFilePath, folderPath, ignore=DEFAULT_IGNORE_KEYWORDS):
        if folderPath in ignore:
            logger.info(
                "Ignore compressing folder: {} -> {}".format(folderPath, zipFilePath)
            )
            return zipFilePath

        logger.info("Compressing folder: {} -> {}".format(folderPath, zipFilePath))
        with zipfile.ZipFile(zipFilePath, "w") as zip:
            for root, _, files in os.walk(folderPath):
                for file in files:
                    filePath = os.path.join(root, file)
                    zip.write(filePath, arcname=self.relativePath(filePath, folderPath))
        logger.info("Compressed folder: {} -> {}".format(folderPath, zipFilePath))
        return zipFilePath

    def compressFile(self, zipFilePath, filePath, ignore=DEFAULT_IGNORE_KEYWORDS):
        if filePath in ignore:
            logger.info(
                "Ignore compressing File: {} -> {}".format(filePath, zipFilePath)
            )
            return zipFilePath

        logger.info("Compressing File: {} -> {}".format(filePath, zipFilePath))
        with zipfile.ZipFile(zipFilePath, "w") as zip:
            _, filename = os.path.split(filePath)
            zip.write(filePath, arcname=filename)
        logger.info("Compressed File: {} -> {}".format(filePath, zipFilePath))
        return zipFilePath

    def extract(self, zipFilePath, distPath, ignore=DEFAULT_IGNORE_KEYWORDS):
        logger.info("Extracting zip: {} -> {}".format(zipFilePath, distPath))
        with zipfile.ZipFile(zipFilePath, "r") as zip:
            zip.extractall(distPath)
        self.removeIgnore(distPath, ignore=ignore)
        logger.info("Extracted zip: {} -> {}".format(zipFilePath, distPath))

    def remove(self, *paths):
        return [path.remove(self.pathJoin(self.tempStore, p)) for p in paths]

    def storageUrl(self, bucket, path):
        return "file://" + path

    def removeIgnore(self, path, ignore=DEFAULT_IGNORE_KEYWORDS):
        for root, folders, files in os.walk(path):
            for folder in folders:
                _path = self.pathJoin(root, folder)
                path.remove(_path)
                logger.info("Removed ignore file: {}".format(_path))
            for file in files:
                _path = self.pathJoin(root, file)
                path.remove(_path)
                logger.info("Removed ignore folder: {}".format(_path))

    def copy(self, path, dist):
        _path = self.pathJoin(self.tempStore, path)
        _dist = self.pathJoin(self.tempStore, dist)
        path.copy(_path, _dist)
        return dist



# coding=utf-8
from __future__ import print_function

import time
import traceback

import redis

from suanpan.log import logger


class RedisMQ(object):
    def __init__(
        self,
        redisHost="localhost",
        redisPort=6379,
        redisRealtime=False,
        options=None,
        client=None,
        **kwargs
    ):
        self.options = options or {}
        self.options.update(host=redisHost, port=redisPort, realtime=redisRealtime)
        if client and not isinstance(client, redis.Redis):
            raise Exception("Invalid client: {}".format(client))
        self.client = client or redis.Redis(
            host=self.options["host"], port=self.options["port"], decode_responses=True
        )

    @property
    def connected(self):
        return bool(self.client.connection)

    def createQueue(self, name, group="default", consumeID="0", force=False):
        if force:
            self.deleteQueue(name)
        return self._createQueue(name, group=group, consumeID=consumeID)

    def _createQueue(self, name, group="default", consumeID="0"):
        try:
            return self.client.xgroup_create(name, group, id=consumeID, mkstream=True)
        except:
            traceback.print_exc()
            raise Exception("Queue existed: {}".format(name))

    def deleteQueue(self, *names):
        return self.client.delete(*names)

    # def hasQueue(self, name, group="default"):
    #     try:
    #         queue = self.client.xinfo_stream(name)
    #         groups = self.client.xinfo_groups(name)
    #         return any(g["name"].decode() == group for g in groups)
    #     except:
    #         return False

    def sendMessage(self, queue, data, messageID="*", maxlen=None):
        return self.client.xadd(queue, data, id=messageID, maxlen=maxlen)

    def recvMessages(
        self,
        queue,
        group="default",
        consumer="unknown",
        noack=False,
        block=True,
        count=1,
        consumeID=">",
    ):
        block = None if not block else 0 if block is True else block
        return self.client.xreadgroup(
            group, consumer, {queue: consumeID}, count=count, block=block, noack=noack
        )

    def subscribeQueue(
        self,
        queue,
        group="default",
        consumer="unknown",
        noack=False,
        block=True,
        count=1,
        consumeID=">",
        delay=1,
        errDelay=1,
        errCallback=print,
    ):
        def _parseMessagesGenerator(messages):
            for message in messages:
                _queue, items = message
                for item in items:
                    _id, _data = item
                    yield {"id": _id, "data": _data, "queue": _queue, "group": group}

        while True:
            try:
                messages = self.recvMessages(
                    queue,
                    group=group,
                    consumer=consumer,
                    noack=noack,
                    block=block,
                    count=count,
                    consumeID=consumeID,
                )
                messages = list(_parseMessagesGenerator(messages))
            except:
                logger.error(traceback.format_exc())
                time.sleep(errDelay)
                continue
            if messages:
                for message in messages:
                    yield message
                self.client.xack(queue, group, *[m["id"] for m in messages])
            else:
                time.sleep(delay)



# coding=utf-8
from __future__ import print_function

import functools

import numpy as np

from suanpan import path
from suanpan.utils import json


def _loads(obj, *args, **kwargs):
    return np.frombuffer(obj["data"], obj["dtype"]).reshape(obj["shape"])


def _load(fp, *args, **kwargs):
    return _loads(json.load(fp, *args, **kwargs))


def _loadf(file, *args, **kwargs):
    encoding = kwargs.pop("encoding", "utf-8")
    with open(file, "w", encoding=encoding) as file:
        return _load(file, *args, **kwargs)


def _dumps(npy, *args, **kwargs):
    return {"dtype": str(npy.dtype), "shape": npy.shape, "data": npy.tobytes()}


def _dump(npy, fp, *args, **kwargs):
    return json.dump(_dumps(npy, *args, **kwargs), fp)


def _dumpf(npy, file, *args, **kwargs):
    encoding = kwargs.pop("encoding", "utf-8")
    path.safeMkdirsForFile(file)
    with open(file, "w", encoding=encoding) as file:
        return _dump(npy, file, *args, **kwargs)


def load(file, *args, **kwargs):
    _l = _loadf if isinstance(file, str) else _load
    return _l(file, *args, **kwargs)


def dump(npy, file, *args, **kwargs):
    _d = _dumpf if isinstance(file, str) else _dump
    return _d(npy, file, *args, **kwargs)


loads = _loads
dumps = _dumps



# coding=utf-8
from __future__ import print_function

import tempfile

from suanpan.arguments import Int, String
from suanpan.proxy import Proxy
from suanpan.storage.local import LocalStorage
from suanpan.storage.oss import OssStorage


class StorageProxy(Proxy):
    MAPPING = {"oss": OssStorage, "local": LocalStorage}
    ARGUMENTS = [
        String("storage-type", default="oss"),
        String("storage-oss-access-id"),
        String("storage-oss-access-key"),
        String("storage-oss-bucket-name", default="suanpan"),
        String("storage-oss-endpoint", default="http://oss-cn-beijing.aliyuncs.com"),
        String("storage-oss-delimiter", default="/"),
        String("storage-oss-temp-store", default=tempfile.gettempdir()),
        Int("storage-oss-download-num-threads", default=1),
        String("storage-oss-download-store-name", default=".py-oss-download"),
        Int("storage-oss-upload-num-threads", default=1),
        String("storage-oss-upload-store-name", default=".py-oss-upload"),
        String("storage-local-temp-store", default=tempfile.gettempdir()),
    ]


storage = StorageProxy()



# coding=utf-8
from __future__ import print_function

from suanpan.arguments import Arg
from suanpan.mstorage import mstorage
from suanpan.utils import json, npy, pickle


class MStorageArg(Arg):
    pass


class Pickle(MStorageArg):
    def load(self, args):
        self.value = super(Pickle, self).load(args)
        return self.value

    def format(self, context):
        if self.value:
            self.value = pickle.loads(mstorage.get(self.value))
        return self.value

    def save(self, context, result):
        obj = result.value
        mstorage.set(self.value, pickle.dumps(obj))
        self.logSaved(self.value)
        return self.value


class Any(Pickle):
    pass


class Npy(MStorageArg):
    def load(self, args):
        self.value = super(Npy, self).load(args)
        return self.value

    def format(self, context):
        if self.value:
            self.value = self.getnpy()
        return self.value

    def save(self, context, result):
        obj = result.value
        self.setnpy(obj)
        self.logSaved(self.value)
        return self.value

    def setnpy(self, obj):
        params = npy.dumps(obj)
        npybytes = params.pop("data")
        data = {"md": json.dumps(params), "data": npybytes}
        return mstorage.mset(self.value, data)

    def getnpy(self):
        data = npy.loads(mstorage.mget(self.value))
        params = json.loads(data["md"])
        params["data"] = data["data"]
        return npy.loads(params)



# coding=utf-8
from __future__ import print_function

import functools
import itertools
import time

from suanpan.log import logger


def cost(func):
    @functools.wraps(func)
    def _dec(*args, **kwargs):
        start = time.time()
        result = func(*args, **kwargs)
        end = time.time()
        cost = round(end - start, 6)
        paramstring = ", ".join(
            itertools.chain(
                (str(a) for a in args),
                ("{}={}".format(k, v) for k, v in kwargs.items()),
            )
        )
        funcstring = "{}({})".format(func.__name__, paramstring)
        logger.info("{} - {}s".format(funcstring, cost))
        return result

    return _dec



# coding=utf-8
from __future__ import print_function

import numpy as np

from suanpan import g, path
from suanpan.arguments import Arg
from suanpan.components import Result
from suanpan.storage import storage
from suanpan.utils import csv


class StorageArg(Arg):
    pass
    # def __init__(self, key, **kwargs):
    #     kwargs.update(default=self.getDefault(kwargs.pop("_keyword")))
    #     super(StorageArg, self).__init__(key, **kwargs)

    # def getDefault(self, keyword):
    #     return storage.pathJoin(g.USER_ID, g.APP_ID, g.NODE_ID, keyword)  # pylint: disable=no-member


class File(StorageArg):
    FILENAME = "file"
    FILETYPE = None

    def __init__(self, key, **kwargs):
        # kwargs.update(required=False)
        fileName = kwargs.pop("name", self.FILENAME)
        fileType = kwargs.pop("type", self.FILETYPE)
        self.fileName = (
            "{}.{}".format(fileName, fileType.lower()) if fileType else fileName
        )
        super(File, self).__init__(key, **kwargs)

    @property
    def isSet(self):
        return True

    def load(self, args):
        objectPrefix = super(File, self).load(args)
        self.objectPrefix = (
            objectPrefix.replace("_", storage.delimiter) if objectPrefix else None
        )
        self.objectName = (
            storage.pathJoin(self.objectPrefix, self.fileName)
            if self.objectPrefix
            else None
        )
        self.filePath = (
            storage.getPathInTempStore(self.objectName) if self.objectName else None
        )
        if self.filePath:
            path.safeMkdirsForFile(self.filePath)
        self.value = self.filePath
        return self.filePath

    def format(self, context):
        if self.filePath:
            storage.download(self.objectName, self.filePath)
        return self.filePath

    def save(self, context, result):
        filePath = result.value
        storage.upload(self.objectName, filePath)
        self.logSaved(self.objectName)
        return self.objectPrefix


class Folder(StorageArg):
    # def __init__(self, key, **kwargs):
    #     kwargs.update(required=False)
    #     super(Folder, self).__init__(key, **kwargs)

    @property
    def isSet(self):
        return True

    def load(self, args):
        folderName = super(Folder, self).load(args)
        self.folderName = (
            folderName.replace("_", storage.delimiter) if folderName else None
        )
        self.folderPath = (
            storage.getPathInTempStore(self.folderName) if self.folderName else None
        )
        if self.folderPath:
            path.safeMkdirs(self.folderPath)
        self.value = self.folderPath
        return self.folderPath

    def format(self, context):
        if self.folderPath:
            storage.download(self.folderName, self.folderPath)
        return self.folderPath

    def clean(self, context):
        if self.folderPath:
            path.empty(self.folderPath)
        return self.folderPath

    def save(self, context, result):
        folderPath = result.value
        storage.upload(self.folderName, folderPath)
        self.logSaved(self.folderName)
        return self.folderName


class Data(File):
    FILENAME = "data"


class Json(Data):
    FILETYPE = "json"


class Csv(Data):
    FILETYPE = "csv"

    def format(self, context):
        super(Csv, self).format(context)
        self.value = csv.load(self.filePath)
        return self.value

    def save(self, context, result):
        csv.dump(self.filePath, result.value)
        return super(Csv, self).save(context, Result(self.filePath))


class Npy(Data):
    FILETYPE = "npy"

    def format(self, context):
        super(Npy, self).format(context)
        self.value = np.load(self.filePath)
        return self.value

    def save(self, context, result):
        np.save(self.filePath, result.value)
        return super(Npy, self).save(context, Result(self.filePath))


class Model(File):
    FILENAME = "model"


class H5Model(Model):
    FILETYPE = "h5"


class Checkpoint(Model):
    FILETYPE = "ckpt"


class JsonModel(Model):
    FILETYPE = "json"



# coding=utf-8
from __future__ import print_function

from suanpan.arguments import Int, String
from suanpan.mstorage.redis import RedisMStorage
from suanpan.proxy import Proxy


class MStorageProxy(Proxy):
    MAPPING = {"redis": RedisMStorage}
    ARGUMENTS = [
        String("mstorage-type", default="redis"),
        String("mstorage-redis-host", default="localhost"),
        Int("mstorage-redis-port", default=6379),
    ]


mstorage = MStorageProxy()
